/*
 *
 *   This file must contain the common controls used by all alorithms
 *
 */
#define __KERNEL_SYSCALLS__             /*  for waitpid */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/net.h>
#include <linux/unistd.h>
#include <asm/unistd.h>
#include <net/sock.h>
#include <net/tcp.h>

#define NO_VERSION
#include "tcpslb.h"

inline extern void  perror(char *str);

/*
 * A proceddure to create a lisetning socket.. returns a non zero on failure..
 * like no resource available.
 * must this be static inline??
 */
int tcpslb_create_listening_sock(struct socket **sock,
		struct tcpslb_service *serv)
{
        struct sockaddr_in mysock;

	/*Create a TCP stream socket*/
        if (sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, sock) < 0) {
           perror("Error during creation of socket; terminating\n");
           return -1;
        }
        memset(&mysock, 0, sizeof(mysock)); //set the structure
        mysock.sin_family = AF_INET;
        mysock.sin_port = htons(serv->port);
        mysock.sin_addr.s_addr = htonl(INADDR_ANY);//must we limit it 
						//to only some interface??
        (*sock)->sk->reuse=1;

	/*Binding to the port*/
      if((*sock)->ops->bind(*sock,(struct sockaddr*)&mysock, sizeof(mysock))<0){
	     perror("bind");
             return -1;
        }   

      /*set up as stream listening socket*/
        if((*sock)->ops->listen(*sock,MAX_WAIT_QUEUE)<0){
	     perror("listen");
             return -1;
        }

	//All fine if we reach here so we return 0
	return 0;
	
}

/*
 * Perform the actual accept of the function...
 * Firstly the blocking call did not work....
 * secondly we could'nt have used one... as while termination the daemon must 
 * be stopped and blocked calls dont die easily!!! 
 * (atleast kill_proc has poor performance... and we cant unload if all threads
 * dont die!!)
 * This proc simulates a blocking call using a non blocking one.. with an added
 * advantage... we needed
 */
int tcpslb_accept(struct socket *listen, struct socket **client, 
		int *tcpslb_cont)
{
        DECLARE_WAITQUEUE(wait, current);

	/* Block all signals except SIGKILL and SIGSTOP */
       spin_lock_irq(&current->sigmask_lock);
       siginitsetinv(&current->blocked,sigmask(SIGKILL) | sigmask(SIGSTOP));
       recalc_sigpending(current);
       spin_unlock_irq(&current->sigmask_lock);

	while(*tcpslb_cont) {      
		/*while we are allowed to continue.
		 * this will be set to 0 by cleanup */
	      if(signal_pending(current)){
	              perror("bailing out"); 
		      return -1; 
		      //some pending signal??
		      //do we need this??? learn more
	      }
	  
	      /*clear the zombies.. Return back the unused space
	       * we expect to clear all groups of zombies w/o hanging*/
	       waitpid(-1, NULL, __WALL|WNOHANG);

	      /*Putting to sleep with timeout if nothing in queue  */
	      if(listen->sk->tp_pinfo.af_tcp.accept_queue == NULL) {
		      add_wait_queue(listen->sk->sleep, &wait);
		      __set_current_state(TASK_INTERRUPTIBLE);
		      schedule_timeout(HZ);
		      __set_current_state(TASK_RUNNING);
		      remove_wait_queue(listen->sk->sleep, &wait);
		      continue;
	      }

	      //we have something in queue... 
	      //we clone the socket.. to accept a new client
	      *client=sock_alloc();         
	      (*client)->type=listen->type;
	      (*client)->ops=listen->ops;

	      //we make a non blocking accept
	      //if we fail we release the new socket and try again
	      if(listen->ops->accept(listen,*client,O_NONBLOCK)<0){
		      sock_release(*client);
		      continue;
	      }
	      return 0; //ALL OK we have accepted a connection
	}
return -2; //we have got out of the main loop.. 
           //cleanup has asked to close the service
}


/* 
 * A procedure to send message... almost enirely stolen from ktcpvs!!
 */
 int
tcpslb_send(struct socket *sock, const char *buffer, const size_t length)
{
        struct msghdr   msg;
        struct iovec    iov;
        int             len;
        mm_segment_t   oldfs;
        iov.iov_base     = (void *)buffer;
        iov.iov_len      = length;
        msg.msg_name     = 0;
        msg.msg_namelen  = 0;
        msg.msg_iov      = &iov;
        msg.msg_iovlen   = 1;
        msg.msg_control  = NULL;
        msg.msg_controllen = 0;
        msg.msg_flags    = MSG_DONTWAIT|MSG_NOSIGNAL;
        oldfs = get_fs(); set_fs(KERNEL_DS);
        len = sock_sendmsg(sock, &msg, (size_t)(length));
        set_fs(oldfs);
        return len;
}

/* 
 * A procedure to send message.. almost stolen ... the only differnce is
 * this implements a  non blocking call.. 
 * We need to find what the flag MSG_PEEK does.. it may be a better option
 */
int
tcpslb_receive(struct socket *sock, const char *buffer, const size_t length)
{
        struct msghdr   msg;
        struct iovec    iov;
        int             len;
        mm_segment_t   oldfs;
        iov.iov_base     = (void *)buffer;
        iov.iov_len      = length;
        msg.msg_name     = 0;
        msg.msg_namelen  = 0;
        msg.msg_iov      = &iov;
        msg.msg_iovlen   = 1;
        msg.msg_control  = NULL;
        msg.msg_controllen = 0;
        msg.msg_flags    = 0;//MSG_DONTWAIT|MSG_NOSIGNAL;
        oldfs = get_fs(); set_fs(KERNEL_DS);
        len = sock_recvmsg(sock, &msg, (size_t)(length),MSG_DONTWAIT);
        set_fs(oldfs);
        return len;
}


/*
 * Simple function to print error... unable to quit habit of app prog
 */
inline void  perror(char *str)
{
	printk("<1> %s",str);
}

