/*
 *                     FIXED HASH SCHEDULING
 */

#define __KERNEL_SYSCALLS__             /*  for waitpid */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/net.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/wait.h>
#include <asm/unistd.h>

#include <net/sock.h>
#include <net/tcp.h>

#include "tcpslb.h"
#include "hash.h"
#include "common.h"


extern __u32 in_aton(const char *str);
int tcpslb_hash_addnode(struct tcpslb_sched_hash *sched,int *addr,
		struct tcpslb_realservers *dest);
static struct tcpslb_realservers *tcpslb_hash_getnode(
		struct tcpslb_hash_node *hash,int *addr);
static int tcpslb_accept_cons_hash(void *service);
static int tcpslb_cross_conn(void *data);
static int tcpslb_serv_other_side(void *data);
static int errno;


static int __init tcpslb_hash_init(void)
{
     inter_module_register("tcpslb_accept_cons_hash", THIS_MODULE, tcpslb_accept_cons_hash);
     inter_module_register("tcpslb_hash_addnode", THIS_MODULE, tcpslb_hash_addnode);
     return 0;
}

static void __exit tcpslb_hash_cleanup(void)
{
  	inter_module_unregister("tcpslb_accept_cons_hash");
     	inter_module_unregister("tcpslb_hash_addnode");
}


int tcpslb_hash_addnode(struct tcpslb_sched_hash *sched,int *addr,
		struct tcpslb_realservers *dest)
{
	struct tcpslb_hash_node *hash,*hash1,*temp,*temp1;
	int i;
	hash1=NULL;
	hash=sched->hash;
	for(i=0;i<4;i++){
		if(hash==NULL){
			hash=(struct tcpslb_hash_node *)vmalloc(
				sizeof(struct tcpslb_hash_node));
			hash->addr=256;      //256 is *
			hash->dest=NULL;
			hash->next=NULL;
			hash->nextLevel=NULL;
			if(i==0){
				sched->hash=hash;
			}
			else{
				hash1->nextLevel=hash;
			}
			
		}
		temp=hash;
		if(addr[i]==256){
			while(temp->addr!=256)	temp=temp->next;
			temp->dest=dest;
			return 0;
		}
	        while(temp!=NULL){
		       	if(temp->addr==addr[i])  break;
			temp=temp->next;
		}

		if(temp==NULL){
			temp=hash;
			temp1=NULL;
			while(temp!=NULL && temp->addr<addr[i]){
				temp1=temp;
				temp=temp->next;
			}
			if(temp1==NULL){
				if(i==0){
 			        sched->hash=(struct tcpslb_hash_node *)
				       vmalloc(sizeof(struct tcpslb_hash_node));
				sched->hash->addr=addr[i];
				sched->hash->nextLevel=NULL;
				sched->hash->next=temp;
				sched->hash->dest=NULL;
				temp=sched->hash;
				}
				else{
 			        hash1->nextLevel=(struct tcpslb_hash_node *)
				       vmalloc(sizeof(struct tcpslb_hash_node));
				hash1->nextLevel->addr=addr[i];
				hash1->nextLevel->nextLevel=NULL;
				hash1->nextLevel->next=temp;
				hash1->nextLevel->dest=NULL;
				temp=hash1->nextLevel;
				}
				
			}
			else{
				temp1->next=(struct tcpslb_hash_node *)vmalloc(
					sizeof(struct tcpslb_hash_node));
				temp1->next->addr=addr[i];
				temp1->next->nextLevel=NULL;
				temp1->next->next=temp;
				temp1->next->dest=NULL;
				temp=temp1->next;
			}
		}
		if(addr[i]==256 || i==3){ 
			temp->dest=dest;
			break; //this makes condition in for useless
		}
		hash1=temp;
		 if(hash1->next!=NULL)
		hash=temp->nextLevel;
	}
	return 0;
}

static struct tcpslb_realservers *tcpslb_hash_getnode(
		struct tcpslb_hash_node *hash,int *addr)
{
	int i;
	struct tcpslb_hash_node *lookback[4];
#ifdef DEBUG_MESSAGE			
	printk(KERN_ALERT "Message from %d.%d.%d.%d",addr[0],
			addr[1],addr[2],addr[3]);
#endif 			
	for(i=0;i<4;i++){
		while(hash->addr!=addr[i] && hash->addr!=256){ 
			hash=hash->next;
#ifdef DEBUG_MESSAGE			
			printk(KERN_ALERT "r1");
#endif			
		}
		if(hash->addr==256 || i==3){
			if(hash->dest!=NULL)
				return hash->dest;
			else{
				for(i=i-1;i>=0;i--){
					hash=lookback[i];
#ifdef DEBUG_MESSAGE			
					printk(KERN_ALERT "lookback");
#endif					
					while(hash->addr!=256) hash=hash->next;
					if(hash->dest!=NULL)
						return hash->dest;
				}
				return NULL;
				
			}
		}
#ifdef DEBUG_MESSAGE			
		printk(KERN_ALERT "got %d",hash->addr);
#endif		
		lookback[i]=hash;
		hash=hash->nextLevel;
		if(hash==NULL){
#ifdef DEBUG_MESSAGE			
		      printk(KERN_ALERT "faced NULL!!");
#endif		      
		      return NULL;
		}
	}
	return NULL;  //if hash is fine we wont reach here!!
}

static int tcpslb_accept_cons_hash(void *service)
{
        struct socket *sock,*sock1;
	int ret;
	struct tcpslb_cons *conn;
	struct tcpslb_service *serv=(struct tcpslb_service *)service;
	struct tcpslb_sched_hash *sched=(struct tcpslb_sched_hash *)serv->data;
	int addr[4];

	//create a listening socket
        if(tcpslb_create_listening_sock(&sock,serv)!=0)	{
		return -1;
	}
	
	//tell the world that the daemon is up
	down_interruptible(&serv->sem); //check ret values if err ret -ERESTARTSYS

	while(1){  //infinite loop to keep accepting conns

		//start acceptin clients
	  if((ret=tcpslb_accept(sock,&sock1,&serv->tcpslb_cont))<0){
#ifdef DEBUG_MESSAGE			
		  perror("accept");
#endif
	          break; //got -2  i.e. cleanup has been triggerd
     	  }

	  conn=(struct tcpslb_cons *) vmalloc(sizeof(struct tcpslb_cons));

	  //fill up the structure... since we have a client waiting for service
	  conn->srcSock  = sock1; //sock1 must be de allocated
	  conn->service  = serv;
	  if(conn->service->scheduling==SCHEDULE_SHASH){
        	  addr[0]=((unsigned char *)&sock1->sk->daddr)[0];
	          addr[1]=((unsigned char *)&sock1->sk->daddr)[1];
        	  addr[2]=((unsigned char *)&sock1->sk->daddr)[2];
	          addr[3]=((unsigned char *)&sock1->sk->daddr)[3];
	  }
	  else {   //obviously scheduling is SCHEDULE_DHASH otherwise
        	  addr[0]=((unsigned char *)&sock1->sk->saddr)[0];
	          addr[1]=((unsigned char *)&sock1->sk->saddr)[1];
        	  addr[2]=((unsigned char *)&sock1->sk->saddr)[2];
	          addr[3]=((unsigned char *)&sock1->sk->saddr)[3];
	  }
	  
	  conn->realServ = tcpslb_hash_getnode(sched->hash,addr);
	  if(conn->realServ==NULL){
		  //we cant accept this entry since no equiv value.. 
		  //in the hash..  is this actually filtering??
		 sock_release(sock1);
	  }
	  else{
		  //we keep threading for each obtained client
		  kernel_thread(&tcpslb_cross_conn,(void *)conn,0);
	  }
	}
	sock_release(sock); //close the listening socket
	up(&serv->sem);//tell cleanup the daemon is returning
	return 0;
}


static int tcpslb_cross_conn(void *data)
{
	
	char *msg;
	int len;
	struct socket *sock;
	struct tcpslb_cons *conn=(struct tcpslb_cons *)data; 
	//this must be deallocated at the end of this function
	static struct sockaddr_in mysock;
	struct tcpslb_sockpair_hash *sockpair;
        
	//a stream socket to real server
	if (sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock) < 0) {
           perror("Error during creation of socket; terminating\n");
           return 0;
        }

	//setting in the values of the server as we know it
	memset(&mysock, 0, sizeof(mysock));
        mysock.sin_family = AF_INET;
        mysock.sin_port = htons(conn->realServ->port);
	//in_aton is giving a warning.. but working fine..
        mysock.sin_addr.s_addr = 
		in_aton(conn->realServ->addr);//addr
#ifdef DEBUG_MESSAGE			
	perror(conn->realServ->addr);
#endif			

	//connect to the real server
       if(sock->ops->connect(sock,(struct sockaddr *)&mysock,
			       sizeof(mysock),0)<0){
		printk("Hit Error!!");
		//if we hit an error ... we dont have to necessarily terminate
		//we can alternatively reschedule to another server  and
		//notify about the lost realserver so we dont connect to 
		//it againe.
		//we must implement that *here* eventually
	        sock_release(sock);
 	        sock_release(conn->srcSock);
	        vfree(conn); 
		return 0;
	}

       //creating the sockpair.. to contain the pair of sockets to be spliced
	sockpair = (struct tcpslb_sockpair_hash *)
		vmalloc(sizeof(struct tcpslb_sockpair_hash));
	//must de alloc this
	sockpair->srcSock=conn->srcSock;
	sockpair->destSock=sock;
	sockpair->cont=1;//this is to continue the execution of both the threads

        //initializing the semaphore... to know status of other side of service
	sema_init(&(sockpair->sem),1);
        //we also need a spin lock to control value of sockpair->cont
	spin_lock_init(&(sockpair->lock));

	//thread again so that we can read @ both sock
        kernel_thread(&tcpslb_serv_other_side,(void *)sockpair,0);
	
	msg=(char*) get_free_page(GFP_KERNEL);

        //we continule till this is 1
	while(sockpair->cont){

		//while the queue is empty and we can continue
		while(skb_queue_empty(&(sockpair->destSock->sk->receive_queue))
				&& sockpair->cont) {

		  /* if we are the socket is not in state TCP_ESTABLSIHED
		   * we must quit by setting sockpair->cont to 0
		   * also if the state is TCP_CLOSE_WAIT and the receive buffer
		   * is empty we must set sockpair->cont to 0 since server has 
		   * closed communication and we are in close state...
		   * 
		   * These statements have some bug!! ... the procedure is 
		   * working fine with telnet (to any port)... but clients like
		   * browsers hold connection with server for a long time
		   */ 
//	              printk("d: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
     		  if(sockpair->destSock->sk->state != TCP_ESTABLISHED && 
		    !(sockpair->destSock->sk->state==TCP_CLOSE_WAIT && 
		    !skb_queue_empty(&sockpair->destSock->sk->receive_queue))){

			  //set the value to 0 and keep it atomic least it be 
			  //modified halfway by some party..
			  spin_lock(&(sockpair->lock));
		  	  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
	  		  break;
		   }
		  /* If we reach here, the queue is empty... 
		   * so we go on interruptible sleep
		   */
		   interruptible_sleep_on_timeout(&sock->wait, HZ);
//	              printk("dw: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;

		//receive data... on queue
		if((len=tcpslb_receive(sockpair->destSock, msg, PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ recv1");
			printk("-%d-",len);
#endif			
	     		break;
		}

		//transmit it
		if(tcpslb_send(sockpair->srcSock,msg,len)<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send1");
#endif
			break;
		}
//	              printk("ds: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
	}
	
	//if we get out of while w/o setting sockpair->cont to zero
	//a send / receive error has occured... closing other side
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	//this side of service is closed... the other side will take some 
	//timeout time to close... to remove this delay we wake up the other
	//side
	wake_up_interruptible(&sockpair->srcSock->wait);

        //checking of other side closed
	down_interruptible(&(sockpair->sem));

	//clear all the resource associated with this connection
	sock_release(sock);
	sock_release(conn->srcSock);
	vfree(conn); 
	vfree(sockpair); //deallocetd here!!!!

	free_page((unsigned long)msg);
	//release the zombie created by service on other side
	 waitpid(-1, NULL, __WNOTHREAD);
	return 0;
	
}

/*
 * This is essential the same funcion as above with the connection to real 
 * server part removed. Only differnce is it acts on opposite sides of the 
 * socket serving the other side..
 */
static int tcpslb_serv_other_side(void *data)
{
	struct tcpslb_sockpair_hash *sockpair = 
		(struct tcpslb_sockpair_hash *)data;
	char   *msg;
	int len;

	//semaphore locked till this thread lives
	down_interruptible(&(sockpair->sem));
	msg=(char*) get_free_page(GFP_KERNEL);
	
	while(sockpair->cont){
		
		while (skb_queue_empty(&(sockpair->srcSock->sk->receive_queue))
				&& sockpair->cont) {
 //		  printk("s: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		
		  if(sockpair->srcSock->sk->state != TCP_ESTABLISHED && 
		   !(sockpair->srcSock->sk->state==TCP_CLOSE_WAIT && 
		   !skb_queue_empty(&sockpair->srcSock->sk->receive_queue))){
			  spin_lock(&(sockpair->lock));
			  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
			  break;
		  }
		  interruptible_sleep_on_timeout(&sockpair->srcSock->wait, HZ);
 //		  printk("sw: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;
	
	     	/* 
		 * Note that this is a non blocking receive..since we have to 
		 * look @ both the received value and also see if can continue
		 * the daemon...
		 */
		if((len=tcpslb_receive(sockpair->srcSock, msg,PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
	     		perror("got out @ recv2");
#endif
			break;
	     	}
	   
		if((len=tcpslb_send(sockpair->destSock, msg,len))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send2");
#endif
			break;
		}
 //		  printk("ss: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
	
	}
	
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	wake_up_interruptible(&sockpair->destSock->wait);
	free_page((unsigned long)msg);

	up(&(sockpair->sem)); 
	//thread terminationg so we clear the semaphore.. telling the 1st side
	//that it may clear up the resources.

	return 0;
}


module_init(tcpslb_hash_init);
module_exit(tcpslb_hash_cleanup);
MODULE_LICENSE("GPL");

