/*
 *  ROUND ROBIN SCHEDULING / WEIGHTED ROUND ROBIN SCHEDULING /
 *  LEAST CONNECTIONS SCHEDULING / WEIGHTED LEAST COST SCHEDULING
 */

#define __KERNEL_SYSCALLS__             /*  for waitpid */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/net.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/wait.h>
#include <asm/unistd.h>

#include <net/sock.h>
#include <net/tcp.h>

/*
 * We dont need some headers... we must dump the few which we dont need
 */
#include "tcpslb.h"
#include "rrlc.h"
#include "common.h"



//can we make most of the variables register??? may be needed for speedy op
//atleaset must keep tcpslb_cont register....??
extern __u32 in_aton(const char *str);
static int errno;  //for errno.h
static int tcpslb_accept_cons_generic(void *service);
static  int tcpslb_cross_conn(void *connection);
static int tcpslb_serv_other_side(void *sockpair);
static struct tcpslb_realservers *tcpslb_next_rr(struct tcpslb_service *serv);
static struct tcpslb_realservers *tcpslb_next_wrr(struct tcpslb_service *serv);
static struct tcpslb_realservers *tcpslb_next_lc(struct tcpslb_service *serv);
static void tcpslb_dec_count_lc(struct tcpslb_cons *conn);
static struct tcpslb_realservers *tcpslb_next_wlc(struct tcpslb_service *serv);
static void tcpslb_dec_count_wlc(struct tcpslb_cons *conn);

/*
 *  Intitializing server here.
 *  A procedure to read a config file and set up the datastructures 
 *  must be called here!!... waiting for lakshmi's Code
 */
static int __init tcpslb_rrlc_init(void)
{
     inter_module_register("tcpslb_accept_cons_generic", THIS_MODULE, tcpslb_accept_cons_generic);
     return 0;
}


/*
 *  This is a generic function to accept connection 1 by one
 */
int tcpslb_accept_cons_generic(void *data)
{
        struct socket *sock,*sock1;
	int ret;
	struct tcpslb_cons *conn;
	struct tcpslb_service *serv=(struct tcpslb_service *)data;

	//create a listening socket
        if(tcpslb_create_listening_sock(&sock,serv)!=0)	{
		return -1;
	}
	
	//tell the world that the daemon is up
	down_interruptible(&serv->sem); //check ret values if err ret -ERESTARTSYS

	
	while(1){  //infinite loop to keep accepting conns

		//start acceptin clients
	  if((ret=tcpslb_accept(sock,&sock1,&serv->tcpslb_cont))<0){
		  perror("accept");
	          break; //got -2  i.e. cleanup has been triggerd
     	  }

	  conn=(struct tcpslb_cons *) vmalloc(sizeof(struct tcpslb_cons));

	  //fill up the structure... since we have a client waiting for service
	  conn->srcSock  = sock1; //sock1 must be de allocated
	  conn->service  = serv;
	  if(serv->scheduling==SCHEDULE_RR)
	        conn->realServ = tcpslb_next_rr(serv);
	  else if(serv->scheduling==SCHEDULE_WRR)
	        conn->realServ = tcpslb_next_wrr(serv);
	  else if(serv->scheduling==SCHEDULE_LC)
	        conn->realServ = tcpslb_next_lc(serv);
	  else if(serv->scheduling==SCHEDULE_WLC)
	        conn->realServ = tcpslb_next_wlc(serv);
	  //write code for rest algs here
	  else{
		  perror("We dont know this alogrithm");
		  break; //it this correct??
	  }
			
	  //we keep threading for each obtained client
          kernel_thread(&tcpslb_cross_conn,(void *)conn,0);

	}
	sock_release(sock); //close the listening socket
	up(&serv->sem);//tell cleanup the daemon is returning
	return 0;
}

/*
 * this Function will open the socket to the real server. and serve one side
 * side of the connection.
 */
static int tcpslb_cross_conn(void *data){
	
	char *msg;
	int len;
	struct socket *sock;
	struct tcpslb_cons *conn=(struct tcpslb_cons *)data; 
	//this must be deallocated at the end of this function
	static struct sockaddr_in mysock;
	struct tcpslb_sockpair *sockpair;
        
	//a stream socket to real server
	if (sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock) < 0) {
           perror("Error during creation of socket; terminating\n");
           return 0;
        }

	//setting in the values of the server as we know it
	memset(&mysock, 0, sizeof(mysock));
        mysock.sin_family = AF_INET;
        mysock.sin_port = htons(conn->realServ->port);
	//in_aton is giving a warning.. but working fine..
        mysock.sin_addr.s_addr = 
		in_aton(conn->realServ->addr);//addr
#ifdef DEBUG_MESSAGE			
	perror(conn->realServ->addr);
#endif

	//connect to the real server
       if(sock->ops->connect(sock,(struct sockaddr *)&mysock,
			       sizeof(mysock),0)<0){
		printk("Hit Error!!");
		//if we hit an error ... we dont have to necessarily terminate
		//we can alternatively reschedule to another server  and
		//notify about the lost realserver so we dont connect to 
		//it againe.
		//we must implement that *here* eventually
		if(conn->service->scheduling==SCHEDULE_LC){
                  tcpslb_dec_count_lc(conn);
		}
		else if(conn->service->scheduling==SCHEDULE_WLC){
                  tcpslb_dec_count_wlc(conn);
		}
	        sock_release(sock);
 	        sock_release(conn->srcSock);
	        vfree(conn); 
		return 0;
	}

       //creating the sockpair.. to contain the pair of sockets to be spliced
	sockpair = (struct tcpslb_sockpair *)
		vmalloc(sizeof(struct tcpslb_sockpair));
	//must de alloc this
	sockpair->srcSock=conn->srcSock;
	sockpair->destSock=sock;
	sockpair->cont=1;//this is to continue the execution of both the threads

        //initializing the semaphore... to know status of other side of service
	sema_init(&(sockpair->sem),1);
        //we also need a spin lock to control value of sockpair->cont
	spin_lock_init(&(sockpair->lock));

	//thread again so that we can read @ both sock
        kernel_thread(&tcpslb_serv_other_side,(void *)sockpair,0);
	
	msg=(char*) get_free_page(GFP_KERNEL);

        //we continule till this is 1
	while(sockpair->cont){

		//while the queue is empty and we can continue
		while(skb_queue_empty(&(sockpair->destSock->sk->receive_queue))
				&& sockpair->cont) {

		  /* if we are the socket is not in state TCP_ESTABLSIHED
		   * we must quit by setting sockpair->cont to 0
		   * also if the state is TCP_CLOSE_WAIT and the receive buffer
		   * is empty we must set sockpair->cont to 0 since server has 
		   * closed communication and we are in close state...
		   * 
		   * These statements have some bug!! ... the procedure is 
		   * working fine with telnet (to any port)... but clients like
		   * browsers hold connection with server for a long time
		   */ 
//	              printk("d: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
     		  if(sockpair->destSock->sk->state != TCP_ESTABLISHED && 
		    !(sockpair->destSock->sk->state==TCP_CLOSE_WAIT && 
		    !skb_queue_empty(&sockpair->destSock->sk->receive_queue))){

			  //set the value to 0 and keep it atomic least it be 
			  //modified halfway by some party..
			  spin_lock(&(sockpair->lock));
		  	  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
	  		  break;
		   }
		  /* If we reach here, the queue is empty... 
		   * so we go on interruptible sleep
		   */
		   interruptible_sleep_on_timeout(&sock->wait, HZ);
//	              printk("dw: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;

		//receive data... on queue
		if((len=tcpslb_receive(sockpair->destSock, msg, PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ recv1");
#endif
	     		break;
		}

		//transmit it
		if(tcpslb_send(sockpair->srcSock,msg,len)<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send1");
#endif
			break;
		}
//	              printk("ds: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
	}
	
	//if we get out of while w/o setting sockpair->cont to zero
	//a send / receive error has occured... closing other side
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	//this side of service is closed... the other side will take some 
	//timeout time to close... to remove this delay we wake up the other
	//side
	wake_up_interruptible(&sockpair->srcSock->wait);

        //checking of other side closed
	down_interruptible(&(sockpair->sem));

        //if this is a LC or WLC connection we need to notify that a
	//connection has been closed
	if(conn->service->scheduling==SCHEDULE_LC){
		tcpslb_dec_count_lc(conn);
	}
	else if(conn->service->scheduling==SCHEDULE_WLC){
		tcpslb_dec_count_wlc(conn);
	}

	//clear all the resource associated with this connection
	sock_release(sock);
	sock_release(conn->srcSock);
	vfree(conn); 
	vfree(sockpair); //deallocetd here!!!!

	free_page((unsigned long)msg);
	//release the zombie created by service on other side
	 waitpid(-1, NULL, __WNOTHREAD);
	return 0;
	
}

/*
 * This is essential the same funcion as above with the connection to real 
 * server part removed. Only differnce is it acts on opposite sides of the 
 * socket serving the other side..
 */
static int tcpslb_serv_other_side(void *data){
	struct tcpslb_sockpair *sockpair = (struct tcpslb_sockpair *)data;
	char   *msg;
	int len;

	//semaphore locked till this thread lives
	down_interruptible(&(sockpair->sem));
	msg=(char*) get_free_page(GFP_KERNEL);
	
	while(sockpair->cont){
		
		while (skb_queue_empty(&(sockpair->srcSock->sk->receive_queue))
				&& sockpair->cont) {
 //		  printk("s: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		
		  if(sockpair->srcSock->sk->state != TCP_ESTABLISHED && 
		   !(sockpair->srcSock->sk->state==TCP_CLOSE_WAIT && 
		   !skb_queue_empty(&sockpair->srcSock->sk->receive_queue))){
			  spin_lock(&(sockpair->lock));
			  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
			  break;
		  }
		  interruptible_sleep_on_timeout(&sockpair->srcSock->wait, HZ);
 //		  printk("sw: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;
	
	     	/* 
		 * Note that this is a non blocking receive..since we have to 
		 * look @ both the received value and also see if can continue
		 * the daemon...
		 */
		if((len=tcpslb_receive(sockpair->srcSock, msg,PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
	     		perror("got out @ recv2");
#endif
			break;
	     	}
	   
		if((len=tcpslb_send(sockpair->destSock, msg,len))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send2");
#endif
			break;
		}
 //		  printk("ss: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
	
	}
	
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	wake_up_interruptible(&sockpair->destSock->wait);
	free_page((unsigned long)msg);

	up(&(sockpair->sem)); 
	//thread terminationg so we clear the semaphore.. telling the 1st side
	//that it may clear up the resources.

	return 0;
}

/* 
 * the cleanup function
 */
static void __exit tcpslb_rrlc_cleanup(void)
{
     inter_module_unregister("tcpslb_accept_cons_generic");

}


/*
 * The round robin scheduler
 */
static struct tcpslb_realservers *tcpslb_next_rr(struct tcpslb_service *serv)
{
	struct tcpslb_sched_rr *sched;

	sched=(struct tcpslb_sched_rr *)serv->data;
	
	if(sched->server==NULL){  
		//this will take care of initialization!
		if(serv->realServ==0){
			printk("Hit Err... we dont have any real servers!!");
	                up(&sched->sem);
			return NULL;
			//we need to do some error handling here!!
			//as we dont have realserv and daemon is running
		}
		else
 		  sched->server = serv->realServ;
	}
	else{
		sched->server=sched->server->next;
		if(sched->server==NULL)
			sched->server=serv->realServ;
	}
	up(&sched->sem);
        return	sched->server;
}

/*
 * The weighted round robin scheduler
 */
static struct tcpslb_realservers *tcpslb_next_wrr(struct tcpslb_service *serv)
{
	struct tcpslb_sched_wrr *sched=(struct tcpslb_sched_wrr *)serv->data;
	struct tcpslb_realservers *server;
	int i,max=0,svr=0;

	down_interruptible(&sched->sem);
	sched=(struct tcpslb_sched_wrr *)serv->data;

	//find the maximum of all servers in history
	//The target server is svr th server
	for(i=0;i<serv->noRealServ;i++){
		if(sched->history[i]>max){
			max=sched->history[i];
			svr=i;
		}
	}

	//if the maximum is 0 then copy weights to history and find the 
	//max again..
	if(max==0){
		for(i=0;i<serv->noRealServ;i++){
			sched->history[i]=sched->weight[i];
			if(sched->history[i]>max){
				max=sched->history[i];
				svr=i;
			}
		}
	}
		
	//we need to schedule the svr th server
	for(i=0,server=serv->realServ;i<svr&&server!=NULL;
			i++,server=server->next);
	sched->history[svr]--;  //decrement the histrory
	up(&sched->sem);
	return server;
}

/*
 * The least connections scheduler
 */
static struct tcpslb_realservers *tcpslb_next_lc(struct tcpslb_service *serv)
{
	struct tcpslb_sched_lc *sched=(struct tcpslb_sched_lc *)serv->data;
	struct tcpslb_realservers *server;
	int i,min=99999,svr=0;
	//min is intialized to incredibly large number
	
	
	//find the server with least connections..
	for(i=0;i<serv->noRealServ;i++){
		if(sched->connections[i]<min){
			svr=i;
			min=sched->connections[i];
		}
	}

	for(i=0,server=serv->realServ;i<svr&&server!=NULL;
			i++,server=server->next);
	down_interruptible(&sched->sem);
	sched->connections[svr]++;  //increment the number of connections	
	
	up(&sched->sem);
	return server;
}

/*
 * A procedure to decrement the number of connections for Least Connections
 */
static void tcpslb_dec_count_lc(struct tcpslb_cons *conn)
{
	int len;
	struct tcpslb_sched_lc *schedLc;
	struct tcpslb_realservers *server;
	len=0;
	schedLc=(struct tcpslb_sched_lc *)conn->service->data;
	server=conn->service->realServ;
	while(server!=conn->realServ || server==NULL){
		len++;
		server=server->next;
	}
	if(server!=NULL){
		down_interruptible(&schedLc->sem);
		schedLc->connections[len]--;
		up(&schedLc->sem);
	}
	
}

/*
 * The weighted least connections scheduler
 * formula: min(ci/wi)  
 * where, ci=current connections to i th server
 *        wi=weight of i th server
 */
static struct tcpslb_realservers *tcpslb_next_wlc(struct tcpslb_service *serv)
{
	struct tcpslb_sched_wlc *sched=(struct tcpslb_sched_wlc *)serv->data;
	struct tcpslb_realservers *server;
	int i,svr=0;
	double min=99999;
	//min is intialized to incredibly large number
	
	
	//find the server with least connections..
	for(i=0;i<serv->noRealServ;i++){
		if(sched->connections[i]/sched->weight[i]<min){
			svr=i;
			min=sched->connections[i]/sched->weight[i];
		}
	}

	for(i=0,server=serv->realServ;i<svr&&server!=NULL;
			i++,server=server->next);
	down_interruptible(&sched->sem);
	sched->connections[svr]++;  //increment the number of connections	
	
	up(&sched->sem);
	return server;
}


/*
 * A procedure to decrement the number of connections for 
 * weighted Least Connections
 */
static void tcpslb_dec_count_wlc(struct tcpslb_cons *conn)
{
	int len;
	struct tcpslb_sched_wlc *schedWlc;
	struct tcpslb_realservers *server;
	len=0;
	schedWlc=(struct tcpslb_sched_wlc *)conn->service->data;
	server=conn->service->realServ;
	while(server!=conn->realServ || server==NULL){
		len++;
		server=server->next;
	}
	if(server!=NULL){
		down_interruptible(&schedWlc->sem);
		schedWlc->connections[len]--;
		up(&schedWlc->sem);
	}
	
}
module_init(tcpslb_rrlc_init);
module_exit(tcpslb_rrlc_cleanup);
MODULE_LICENSE("GPL");

